\     Turtle-Logo in Forth, Copyright (c) 2016-2020, Assad Ebrahim, Mathematical Science & Technologies
\          FREE download from: http://www.mathscitech.org/articles/turtle-logo-forth
\          Email assadebrahim2000@gmail.com for comments, bug-reports, feature-requests

\     GNU License - Users are permitted free re-use provided:                          
\           1/ Source code for all modified versions are made freely available under the GNU License            
\           2/ Attribution and this header is preserved (lines 1-7)

\     Description: This is a Turtle logo program operated entirely with keys, for use in Windows .  
\                  It is written in c.1000 (970) lines of Forth code for F-PC Forth and runs within vDOSplus, 
\                       a freely available DOS emulator included in this package.
\                  Windows installer is freely available for download from website:
\                        http://www.mathscitech.org/articles/turtle-logo-forth
\                       and installs to:     C:\TOTALCMD\TURTLE-LOGO\
\                  The program is engaging for children from 3 years onward, and includes advanced features such as 
\                       macro recording and replay, which are engaging for children from 8 years onwards.

\     Usage:        Run TurtleLogo.exe
\     Developers:   Change turtle-autoexec.txt uncommenting last line and comment previous two lines.  hen run TurtleLogo.exe
\                   This will start F-PC
\                   To modify the program, edit turtle.fs and run it in F-PC (Forth FPC) using command:
\                                   fload turtle.fs, then: turtle-go
\                   WARNING:  F-PC does not like TAB in definitions.  Do not use.
\                   To package a modified release, recompile F-PC with FMETA to create F-PC with turtle.fs embedded in it.

\ === REVISION HISTORY
: rev# ." 1.8.4054" ;
\ === INSTRUCTIONS
: welcome ( -- )
    ." WELCOME TO TURTLE LOGO! (v" REV# ." ), (C) 2016-2020, ASSAD EBRAHIM"              \ All Rights Reserved
    cr ."   (1) MOVING TURTLE: use arrow keys. Screen edges wrap to other side!"
    cr ."     [FWD] moves one step along current heading & paints pixel behind Turtle."
    cr ."     [LEFT]/[RIGHT] turns Turtle's heading 45 deg counter-/clock- wise."
    cr ."     [BACKSPACE] backs up & erases pixel (white paint)"                   
    cr ."     [H] sends Turtle to home position at bottom left of screen."
    cr ."   (2) TURTLE'S DRAWING MODES: "
    cr ."     [C/X] cycles pen through 16 color choices (fwd/back)."
    cr ."     ['] toggles pen-up to move without painting; pen-down to draw again."
    cr ."     [ puts Turtle into text mode (now type)   ] returns Turtle to draw mode."
    cr ."     [;] toggle (hide/show) Turtle's visibility  [Z] toggle cloak (stealth) mode"  \ MAX WIDTH OF SCREEN
    cr ."   (3) PROGRAMMING TURTLE: Turtle can learns skills that you TEACH it!"       \ RECORD/REPLAY macros..."
    cr ."     [T] starts/ends TEACHING Turtle a SKILL. (Turtle can learn 9 skills.)"   \ skill=diary.  can record nested macros, but not recursive macros
    cr ."     [N] cycles through skills 1-9."                                         \ Note, it will be EXACTLY what you told it.
    cr ."     [1-9] Turtle executes selected skill. [0] lists all skills known."
    cr ."   (4) OTHER COMMANDS:"
    cr ."     [ESC] clears screen (once you confirm). "
    cr ."     [S] saves the screen; [U] undoes all actions to the last save point"
    cr ."     [#] loads screen from last saved file or previous session"
    cr ."     [G] shows grid coords of Turtle; [W] wave! (Turtle waves back!)"
    cr ."     [?] shows this HELP screen (your drawing is safe!)"
    cr ."     [Q] quits the program (after confirming)"
    cr ." HAPPY TURTLE-ING!"
\    cr ." For more downloads: http://www.mathscitech.org/articles/turtle-logo-forth"
    cr ;

\ === Interactivity
  $1B  constant  esc-key          \ 27d
  $C8  constant  fwd-arrow
  $CD  constant  right-arrow
  $CB  constant  left-arrow
  $8   constant  backspace
  $27  constant  tick            
: pause-key  ( -- )     ." PRESS ANY KEY TO CONTINUE... " key drop ;
: pause-esc  ( -- t/f ) ." PRESS any key for next skill; ESC to return to drawing..." key esc-key =  ;

\ === Bit manipulation
: shl4      ( val -- val )        $10 * ;             \ raise nybble
: shr4      ( val -- val )        $10 / ;             \ lower nybble

\ === Memory Utilities
: bytes     ( n -- n )  ;                                     \ 1x, for documentation
: cells     ( num -- num*wsz )        2 *  ;                  \ 16-bit word length for DOS (80x86 chip)
: c@+       ( adr -- adr' val )       dup 1 + swap c@ ;       \ fetch/store char/word and increment pointer
: @+        ( adr -- adr' val )       dup 1 cells + swap @ ;
: c!+       ( val adr -- adr' )       dup 1 + -rot c! ;
: !+        ( val adr -- adr' )       dup 1 cells + -rot ! ;
: cmove+    ( c-adr dest n -- dest' )  \ copy byte sequence/string (c-addr, n bytes) to dest adr, returning new position
    >r tuck r@ cmove r> + ;     
: attach+        ( dest-adr c-adr nchar -- dest-adr' )   >r swap r> cmove+ ;  \ attach to next memory point, returning new position
: get-fixedwidth-string  ( base-adr fw-len idx -- c-adr fw-len ) over * rot + swap ;   \ returns idx-th fixed width string

\ === Character & Number Utilities
: invert      ( n -- n' )       not   ;                             \ bitwise logical not (invert odd)
: is_between? ( n a b -- t/f )  rot dup rot > -rot > or invert ;    \ not ( n>b or n<a )
: is_char?    ( asc -- t/f )    $20  $7E is_between? ;
: tolower     ( char -- low-char )    $20 or ;                      \ make lowercase
: toupper     ( char -- upp-case )    $20 invert and ;              \ make uppercase
: toggle_case ( char -- tog-case )    $20 xor ;                     \ switch case
: a2d         ( asc -- digit )  '0' - ;                             \ ascii to digit
: d2a         ( digit -- asc )  '0' + ;                             \ digit to ascii
: 2d2a        ( 2dnum -- tens units ) 10 /mod d2a swap d2a swap ;   \ 2 digit number to 2 letter ascii

\ === Pattern matching against a specified keyset, automatically checking lower and uppercase keys
: check-key ( keyn keyp -- keyp t/f )    \ tests if key matches lower/upper case keyn
    swap tolower dup toupper rot swap over = -rot swap over = rot or ;
: check-keyset ( k1 .. kn n keyp -- keyp t/f ) \ testing through the list of n patterns k1 k2 ... (cumulative or)
    swap 0 swap over  ( k1 k2 keyp t/f )  do  -rot check-key rot or  loop    swap drop ;

\ === Strings
$22   constant   dquote    \ double quote
: s"  ( <inputstream"> -- c-adr nchar )  dquote word c@+ pad swap 2dup >r >r cmove r> r> ;   \ quote-terminated str to quoted str
: q2c-string ( c-adr nlen -- ) dup c, here swap dup bytes allot cmove ;   \ quoted string to counted string
    
\ === File Utilities (for F-PC)
: file-write   ( addr len filename --- )   \  write len bytes from addr into the file specified by counted string filename
    count seqhandle+ ">handle ( -- addr len )
    seqhandle+ HDELETE   DROP ( -- )
    seqhandle+ HCREATE   ABORT" Save Create ERR!" ( -- )
    seqhandle+ HWRITE ( addr len -- nwrite ) 0= ABORT" Save Write  ERR!"
    seqhandle+ HCLOSE    ABORT" Save Close  ERR!"  ( -- )  ;
: file-read ( -- )
    count seqhandle+ ">handle
    seqhandle+ HOPEN DROP ( -- )
    seqhandle+ HREAD ( adr len -- nread ) 0= ABORT" Read Savefile ERR!"
    seqhandle+ HCLOSE  ABORT" Read Close ERR!" ( -- )  ; 

\ Filenames - program writes screen & diary memory to file when user saves [s] or quits [q] program
create fscr-save s" save.scr"  q2c-string
create fscr-quit s" quit.scr"  q2c-string
create fskl-save s" save.skl"  q2c-string
create fskl-quit s" quit.skl"  q2c-string

\ === COLOR / PEN MANAGEMENT:   
\    Key concepts: + pen paints background color (bgcolor aka paint-color)
\                  + turtle symbol is foreground color (fgcolor aka turtle-color) 
\                  + user selects background color
\                  + system selects optimal foreground color to contrast with background enough to discern turtle 
\                       while not confusing user as to which paint-color (bgcolor) turtle will draw when it moves.
\    Metric: Contrast Ratio from contrast checker tool, part of theory of color contrasts
\               http://www.worqx.com/color/color_contrast.htm
\               http://webaim.org/resources/contrastchecker/
\               http://juicystudio.com/services/luminositycontrastratio.php#specify
\               Started with this theory, then adjusted to suit personal/user tested impression

\ =========  16 Colors in this video mode (0..15)  Each color is given a string name shown to user
16               constant   num-colors
num-colors 1 +   constant   num-color-strings    \ colors plus penup string
    0      constant  black       1      constant  navy      2      constant  green
    3      constant  teal        4      constant  red       5      constant  purple 
    6      constant  brown       7      constant  silver    8      constant  slate 
    9      constant  blue       10      constant  lime     11      constant  aqua  
    12     constant  orange     13      constant  pink     14      constant  yellow
    15     constant  white 
 \ alternate spellings
    silver constant  grey        silver constant  gray       
    slate  constant  darkgrey    slate  constant  darkgray

6            constant  color-strlen
num-colors   constant   penup-string         \ penup string sits in last index
: create-color-names
    create  here  num-color-strings  color-strlen  *    bytes allot
      " Black " attach+      " DkBlue" attach+      " Green " attach+  \ navy=dark blue
      " Teal  " attach+      " Red   " attach+      " Purple" attach+  \ teal=aqua,  red=maroon
      " Brown " attach+      " Grey  " attach+      " Slate " attach+  \ silver=grey, slate=dark gray
      " Blue  " attach+      " Lime  " attach+      " Aqua  " attach+  \ Blue=light blue, lime=light green, aqua=cyan/lt aqua
      " Orange" attach+      " Pink  " attach+      " Yellow" attach+  \ pink=magenta/fuschia
      " White " attach+      " PEN-UP" attach+
        drop
    does> ( idx base-adr -- c-adr nchar )  \ returns fixed-width color-name
        color-strlen rot get-fixedwidth-string ;   
create-color-names LOOKUP-COLOR-NAME

\ == COLOR VARIABLES 
create border-color 1 bytes allot 
: set-border-color ( color -- ) border-color c! ;
: get-border-color ( -- color ) border-color c@ ;

create canvas-color 1 bytes allot
: set-canvas-color ( color -- )   canvas-color c! ;
: get-canvas-color ( -- color )   canvas-color c@ ;

\ === Color code  (16-color text mode) is one byte: background color in hi nybble, foreground color in low nybble
variable local-fgbg   \  for repetitive paint jobs: fill-screen, border-painting, string writing, menu painting
: fgbg       ( fgcolor bgcolor -- colorcode )  shl4 + ;
: cc2fgbg    ( colorcode -- fgcolor bgcolor )  dup $0F and  ( fgcolor )  swap shr4  ( bgcolor ) ;
: set-fgbg   ( fgbg -- ) local-fgbg ! ;
: get-fgbg   ( -- fgbg ) local-fgbg @ ;
: pure-color ( color -- fgbg ) dup fgbg ;
lime pure-color set-fgbg

: fg-lookup  \ fg-lookup indexed on bgcolor (user selected paint color)
create
 \ fg               bg
   silver   c,    \ black    
   aqua     c,    \ navy     
   lime     c,    \ green
   aqua     c,    \ teal     
   pink     c,    \ red      
   white    c,    \ purple   
   lime     c,    \ brown    
   slate    c,    \ silver   
   silver   c,    \ slate    
   aqua     c,    \ blue     
   green    c,    \ lime     
   navy     c,    \ aqua     
   yellow   c,    \ orange   
   red      c,    \ pink     
   orange   c,    \ yellow   
   silver   c,    \ white     
   \ Penup is handled by SET-PENUP-COLOR
does> ( bg -- fgcolor ) + c@ ;
fg-lookup  LOOKUP-TURTLE-COLOR

create turtle-color      1 bytes allot   \ turtle symbol color is foreground color, chosen by algorithm based on paint (bgcolor)
create paint-color       1 bytes allot   \ this is the color that turtle draws (bgcolor)
create saved-paint-color 1 bytes allot   \ needed so pen-down can restore previous pen-color
: set-turtle-color    ( bgcolor -- ) turtle-color c! ;
: set-paint-color     ( c -- )  dup  paint-color c!  lookup-turtle-color set-turtle-color ; \ auto-sets turtle-color so can be seen
: set-penup-color     ( -- )    lime set-paint-color white turtle-color c! ;   \ gives penup a distinct appearance
: get-turtle-color    ( -- bgcolor ) turtle-color c@ ;
: get-paint-color     ( -- c )       paint-color  c@ ;
: save-paint-color    ( -- )         get-paint-color  saved-paint-color c! ;
: restore-paint-color ( -- c )                        saved-paint-color c@ set-paint-color ;
: change-color        ( 1/-1 -- )    get-paint-color  +  16  mod  set-paint-color ;  \ +1/-1 cycle fwd/bk to next color of 16

\ TURTLE PEN-MODE 
create penup-state          1 cells allot  \ TRUE if pen is up else false
create saved-penup-state    1 cells allot  \ needed so uncloaking does not clobber penup state
: is-penup?           ( -- t/f )           penup-state c@  ; 
: set-penup           ( -- )     true      penup-state c!  ; 
: clear-penup         ( -- )     false     penup-state c!  ; 
: save-penup-state    ( -- )     is-penup? saved-penup-state c! ;  
: restore-penup-state ( -- )               saved-penup-state c@ penup-state c! ;
: get-pencolor-name     ( -- c-adr nchar )  is-penup?  if penup-string else get-paint-color then lookup-color-name ;

\ =========== Graphics.
\ Graphics in DOS is easier than in Windows because DOS allows accessing screen pixels (peek/poke) directly via VGA card memory.  
\ In Windows after Win98 all access to graphics devices requires going through a graphics library such as SDL or OpenGL.
\ F-PC (a 16-bit Forth) is chosen because it runs on DOS and permits 8086 assembly, which make VGA graphics straightforward. 
\ F-PC runs on modern Windows through a DOS emulator such as vDosPlus (free, recommended) or DOSBox.

\ Our graphic protocol is 80x25 text mode graphics, i.e. 2000 pixels (80 columns x 25 rows, indices x:0..79, y:0..24 ).  
\ Behind the scenes there are actually 160 data columns, so 4000 cells, but only the even columns (0,2,...,158) actually display.
\ Pixel locations are given in linear order from 0 to 3998d ($F9E in hex) with final 3999d cell an odd cell so not visible per above.
\   References:
\   Video programming:  http://ece.wpi.edu/~wrm/Courses/EE3803/Labs/roehrl.html
\   Video mode:         http://www.wagemakers.be/english/doc/vga

\ Coordinate representation for computer graphics (our application) is modified Cartesian, i.e. 
\     x,y as across and down but with (0,0) in top left of screen, so y is positive going down..
\ This is different from matrix representation which shares (0,0) as top left of matrix, but i,j coordinates are reversed: 
\     i,j are down (row) then across (column).

\ TWO coordinate systems and their mappings to the underlying implementation (array of memory mapped in row-major form): 
\     DISPLAY/PHYSICAL COORDINATES - where action happens on screen, ALL 160 columns, including non-visible odd cols
\     USER COORDINATES are what user can access on the screen and exclude menus and borders and only shows the VISIBLE 80 columns
\           0,0 as origin point for user's cursor in top left of paintable screen area (subset of 80x25 full graphics area)

\  IMPLEMENTATION: is as an array of memory mapped in row-major convention, i.e. adjacent horizontally are adjacent in the index
\     adjacent vertically are separated by N, number of cols
\     LINEAR INDEX - the underlying implementation / linear data array against which the drawing pixels must be referenced,
\           including the odd invisible pixels issue
\   LIND converts from DISPLAY (PHYSICAL) coordinates to LINEAR INDEX.  No need for back-conversion.
\   LIN converts from USER coordinates to LINEAR INDEX  
\   ILIN converts from LINEAR INDEX to USER coordinates.

\ Example - FPC uses rows 0 and 1 (row 0 is Forth bar, and row 1 erased after command scrolls screen up)
\     To see a pixel within F-PC start at linear position 320 (start of graphics row 3) which is the 0,0 of user coordinates.
\ FPC uses columns 69-79 of the first 12 rows for dictionary, with the full 80 cols accessible from row 13.

\ PC Video Mode (BIOS settings)
$1  constant text40x25-8color
$3  constant text80x25-8color            \ text mode (8 color)
$8  constant text80x25-16color           \ *TURTLE LOGO (and default F-PC) mode, character is 8x8 pixels (640 pixels x 200)
$4  constant graphCGA-320x200-8color     \ standard resolution (64K pixels, 40x25 characters)
$10 constant graphEGA-640x350-16color    \ high resolution (224K pixels, 80x43 characters)
$12 constant graphVGA-640x480-16color    \ default VGA graphics mode (307K pixels, 80x60 characters)

code _set-video-mode ( mode -- )   \ calls PC BIOS instruction to set desired video mode
    pop ax          \ AL=video mode 
    mov ah, # 0     \ AH=set video mode (BIOS instruction)
    int $10         \ DOS interrupt: call BIOS instruction
next end-code

: text-video-mode     ( -- )  text80x25-16color        _set-video-mode ;
: graphics-video-mode ( -- )  graphVGA-640x480-16color _set-video-mode ;   \ F-PC doesn't work in this mode

\ === BASE GRAPHICS - Painting to Screen- Low-level Character Routines
\ color information is encoded into a single byte for use in assembly language routine.  
\ color-code (e.g. $F4) holds background color (bgcolor) in high nybble ($F is white)
\                              foreground (text) color in low nybble ($4 is red)
\ pixel position is in linear encoded form, i.e. row/col are encoded within a single wrap-around array of screen positions

code [poke-pixel]  ( linear-location  ascii-code  color-code  -- )  
        \ writes pixel to screen via direct memory address on vga card
        \ pixel is at physical position in linear array with given colors and ascii character.
        \ Examples: 
        \    320  'X' $7F [poke-pixel]  \ row 2, first column
        \    320  'X' 228 [poke-pixel]
        \    3998 'X' 228 [poke-pixel]  \ row 25, last column
    pop ax                  \ load AL with color-code (foreground & background)
    pop cx                  \ load CL with ascii char
    mov ch, al              \ complete CX as text pixel: (color|char) in (CH|CL)
    pop bx                  \ load BX with screen location
    push ds                 \ preserve data stack
    mov ax, # $B800         \ load AX with graphics address B800 hex (vga memory base address for direct display to screen)
    mov ds, ax              \ set data segment to graphics address (cannot load literal directly, only via register)
    mov 0 [bx], cx          \ display pixel in desired location in graphics memory [B800:bx] (0 data segment offset)
    pop ds                  \ restore data stack
next end-code

: poke-pixel ( linloc char -- ) get-fgbg [poke-pixel] ;  \ wrapper simplifying since color is changed less frequently

code peek-pixel   ( linear-location -- ascii-code color-code )  \ get pixel info from screen; if position is x y use lin first
\ test: turtle-init 0 20 lin peek-pixel
    pop bx              \ load BL with screen location
    push ds             \ preserve data stack
    mov ax, # $B800     \ load AX with graphics address (B800 hex)
    mov ds, ax          \ set data segment to graphics address
    mov cx, 0 [bx]      \ load CX with pixel data from desired location in graphics memory (B800:bx)  (CH|CL)=(color|char)
    pop ds              \ restore data stack
    mov ax, # 0         \ clear AX
    mov al, ch          \ load AX with color-code
    mov ch, # 0         \ clear CH, leaving CL with ascii code
    push cx             \ return ascii code    
    push ax             \ return color-code
next end-code

\ === SCREEN GEOMETRY
  25 constant rows          \ m rows indexed by y or i
  80 constant cols          \ n columns indexed by x or j    (visible columns)
 \  160 constant physical-cols  \ odd columns are invalid display locations, so counter runs 0,2,4,..,178, hence 2 * cols
  2  constant evens-only    \ odd columns are invalid display locations, so counter runs 0,2,4,..,178  
                            \ cols (0..79) to even x-slots (0,2,4,...158) since only evens are visible
   cols rows *    constant   num-pix-scr
\ === LINEAR INDEX: 80x25 (=2000) paintable cells (the evens) out of 160x25 (=4000) cells.
\     0  constant physical-start
\  3998  constant physical-end    \ $F9E hex
\ DESIGN PARAMETERS (independent of coordinate system
  1  constant frame-top       \ frame around user screen
  1  constant frame-left
  1  constant frame-bot
  1  constant frame-right
  1  constant top-menu-rows
  3  constant bottom-menu-rows
  
\ === CONVERSION FROM COORDINATES to LINEAR INDEXING (display implementation)
\ DISPLAY (PHYSICAL) coords (without menu, no frame, etc.) to LINEAR INDEX
: lind ( xcol yrow -- lin )  cols *   +   2 * ;   \ lind = 2*(yrow*80+xcol) = evens_only*(yrow*cols+xcol)
\ USER coordinates (avoids menu, frames) to LINEAR INDEX
: lin ( xcol yrow -- lin )   frame-top top-menu-rows + +  cols *    +    frame-left +    evens-only * ;  
    \ lin = 2*((yrow+1)*80 + (xcol+1))  = evens_only*((yrow+frame_top)*cols + (xcol+frame_left))
    \ test: 0 0 lin  = 320=$140, is 3rd row down.  (1st is menu, 2nd is border)
: ilin  ( lin -- x y )                      \ invert LINEAR INDEX to USER coordinates
    2 /  cols /mod ( x y ) 1 - swap 1 - swap ;

\ DISPLAY COORDINATES
top-menu-rows                 constant top-border-row  \ user coordinates start here in display coordinates
rows bottom-menu-rows - 1 -   constant bot-border-row

\ IN USER COORDINATES
\ screen wrap-around (wrap-y needs an offset when have a top-menu)
rows  top-menu-rows frame-top frame-bot bottom-menu-rows + + + -  constant  user-rows
cols  frame-left frame-right + -                                  constant  user-cols
: wrap-y ( y -- y* )  user-rows mod ;
: wrap-x ( x -- x* )  user-cols mod ;
: init-turtle-position ( -- x y ) 0 user-rows 1 - ;

\ === PAINTING == 
\ displaying strings to any location on screen  (see CARTESIAN for example dynamic string creation)
: write-char+ ( linloc c-adr  -- linloc' c-adr' )  \ writes character on screen position using color code stored in local-fgbg
                                                   \ ex: red lime fgbg set-fgbg  0 20 lin  s" Hello!" drop write-char+
        c@+ rot dup 2 + swap rot poke-pixel swap ; \ next location is 2 bytes ahead - 160 physical positions but 80 screen positions
: write-string ( linloc c-adr nchar -- )   \ ex: white green fgbg set-fgbg  0 20 lin  s" Hello!" write-string
     0  do write-char+ loop 2drop ;

\ PAINTING in DISPLAY COORDINATES
: fill-screen ( color -- )  \  save color then fill even pixels with colored blanks.  (odd pixels go green in this video mode.)
    pure-color set-fgbg
    num-pix-scr 0 do i evens-only * BL poke-pixel loop ;
: black-screen  ( -- ) black fill-screen ;
: vline-draw ( x y1 y2 color -- ) \ draw vertical line at x from y1 to y2 of color
    pure-color set-fgbg  
    swap do ( x ) dup i LIND bl poke-pixel loop drop ;   
: hline-draw ( x1 x2 y color -- ) \ draw horizontal line at y from x1 to x2 of color
    pure-color set-fgbg
    -rot swap do ( y ) i over LIND  bl poke-pixel loop drop ;   
: borders-draw ( -- )  \ all coords are in physical space
    0    top-border-row   bot-border-row    get-border-color   vline-draw
    79   top-border-row   bot-border-row    get-border-color   vline-draw 
    0    80               top-border-row    get-border-color   hline-draw
    0    80               bot-border-row    get-border-color   hline-draw ;

\ === PIXELS & PICTURE
\ A pixel is a 2-byte pair of ascii code and color-codes stored in that order.
\ Picture is a single array (nrow x ncol x 2 bytes) of 2-byte pixels used to save/restore picture to/from memory.
\ Pixel variable is temp storage holding 1 pixel information while user is deciding, in case picture gets impacted (home, penup/down)
2 constant pixel-sz
create pixel     pixel-sz bytes allot    \ storage for 1 pixel
num-pix-scr pixel-sz *       constant   pic-sz
: pix!+ ( ascii-code color-code pix-ptr -- pix-ptr' ) >r swap r> c!+ c!+ ;       \ store pixel, preserving pixel order
: pix@+ ( pix-ptr -- ascii-code color-code pixl-ptr' ) c@+ swap c@+ swap ;       \ fetch pixel

\ memory blocks for system and user to preserve/restore graphics independently
create  system-picture    pic-sz  bytes allot         
create  user-picture      pic-sz  bytes allot
: picture-clear ( pic-ptr --- )  pic-sz 0 fill ;            \ zero out memory
system-picture  picture-clear 
user-picture    picture-clear


\ === DRAWING STATE VARIABLES   
create cloak-state 1 cells allot  \  TRUE allows cloaking (movement without detection, i.e. invisibility mode)
: enable-cloaking  ( -- )     true  cloak-state c! ; 
: disable-cloaking ( -- )     false cloak-state c! ; 
: is-cloaking-on?  ( -- t/f )       cloak-state c@ ; 
create visible-state 1 cells allot  \ Show/Hide Turtle  turtle visibility state -- TRUE if visible, false if hidden
: set-visible ( -- )    true visible-state c! ;
: set-hidden  ( -- )    false visible-state c! ;
: is-visible? ( -- t/f)       visible-state c@ ;

\ === Turtle Penmode changes
: no-action-needed ( -- ) ; 
: cloak-check ( -- ) is-cloaking-on? if no-action-needed else set-visible then ;
: pen-up     ( -- ) \ user requests to put pen-up.  pen-up mode has custom appearance;
    is-penup? if no-action-needed else  save-paint-color  set-penup-color  set-penup  then ; 
: pen-down       ( -- ) clear-penup  restore-paint-color  cloak-check ;
: penup-reset    ( -- ) is-penup? if pen-up else pen-down then ; 
: penup-toggle   ( -- ) is-penup? if pen-down else pen-up then ; 

\ Turtle's Text mode
create text-mode 1 cells allot     \ TRUE if in text mode else FALSE in draw mode
: set-text-mode ( -- )     true  text-mode c! ;
: draw-mode     ( -- )     false text-mode c! ; 
: is-text-mode? ( -- t/f )       text-mode c@ ; 

\ Screenshot pixel by pixel.  Each pixel is a byte in memory, but occupies two bytes on screen.
: screen-save  ( pic-ptr -- ) \ save: screen to memory
    num-pix-scr 0 do i evens-only * ( pic linloc ) peek-pixel rot pix!+ loop drop ;
: screen-restore  ( pic-ptr -- ) \ load: memory to screen
    num-pix-scr 0 do i evens-only * swap pix@+ >r [poke-pixel] r> loop drop  ;
: screen-store ( pic-ptr filename -- ) pic-sz  swap file-write ;  \ save: screen to file (actually mem2file)
: screen-load ( pic-ptr filename -- )  over swap pen-up pic-sz swap file-read screen-restore ;  \ load: file to screen

\ === TURTLE MOVEMENT CONTROL
\ movement is either forward (paints) or backwards (erases) along the axis of the current heading, with screen wrap-around
\ coordinates (x,y) are over and down (not Cartesian).  (0,0) is top left of user space excluding screen border and menus.
\ movement sequence is: 1) update coords to new pixel, 2) paint or restore previous pixel, 3) new pixel is cached, 4) show turtle in new pixel.

\ movement along direction axes
: move-N-fwd   ( x y -- x' y' ) 1 - wrap-y ;
: move-N-back  ( x y -- x' y' ) 1 + wrap-y ;
: move-E-fwd   ( x y -- x' y' ) swap 1 + wrap-x swap ;
: move-E-back  ( x y -- x' y' ) swap 1 - wrap-x swap ;
: move-S-fwd   ( x y -- x' y' ) move-N-back ;
: move-S-back  ( x y -- x' y' ) move-N-fwd ;
: move-W-fwd   ( x y -- x' y' ) move-E-back ;
: move-W-back  ( x y -- x' y' ) move-E-fwd ;
: move-NE-fwd  ( x y -- x' y' ) move-N-fwd move-E-fwd ;
: move-NE-back ( x y -- x' y' ) move-N-back move-E-back ;
: move-SW-fwd  ( x y -- x' y' ) move-NE-back ;
: move-SW-back ( x y -- x' y' ) move-NE-fwd ;
: move-NW-fwd  ( x y -- x' y' ) move-N-fwd move-W-fwd ;
: move-NW-back ( x y -- x' y' ) move-N-back move-W-back ;
: move-SE-fwd  ( x y -- x' y' ) move-NW-back ;
: move-SE-back ( x y -- x' y' ) move-NW-fwd ;

\ === Heading through 8 Compass Directions
\ Heading is an integer 0 - 7 which indexes into an array of heading symbols that show which direction turtle is heading
\ Turtle cycles mod-8 through the directions as cw (turning right)  N > NE > E > SE > S > SW > W > NW  or ccw (turning left)
create heading 1 bytes allot  \  VARIABLE - holds turtle heading, integer 0 to 7    
: set-heading  ( h# -- )    heading c! ;
: get-heading  ( -- h# )    heading c@ ;
: init-heading ( -- )    0  set-heading ;  \ default is N=0
: create-heading-symbol
create ( <name> -- ) '^' c, '/' c, '>' c, '\' c, 'v' c, '/' c, '<' c, '\' c,   \ in cw cycle order (N,NE,E,...)
    does> ( h# -- asc )  + c@ ;
create-heading-symbol  LOOKUP-HEADING-SYMBOL
: create-heading-string
    create ( <name> -- ) here 8 cells allot
    " N " attach+  " NE" attach+   " E " attach+   " SE" attach+
    " S " attach+  " SW" attach+   " W " attach+   " NW" attach+
            drop
    does> ( h# base-adr -- c-adr nchar )  \ returns fixed width direction heading
        2 rot get-fixedwidth-string ;
create-heading-string  LOOKUP-HEADING-STRING

\ HEADING BASED MOTIONS
: create-fwd-motions   \ create motion array of movements in heading-index (rotation cycle) order
    create ( | <name> -- )     
    ['] move-N-fwd   ,   ['] move-NE-fwd  ,    ['] move-E-fwd   ,   ['] move-SE-fwd  ,  
    ['] move-S-fwd   ,   ['] move-SW-fwd  ,    ['] move-W-fwd   ,   ['] move-NW-fwd  , 
    does> ( x y -- x' y' ) 
        get-heading cells + @ execute ;
create-fwd-motions    MOVE-FWD-BY-HEADING   ( x y -- x' y' )
: create-back-motions
    create ( | <name> -- )     
    ['] move-N-back   ,   ['] move-NE-back  ,    ['] move-E-back   ,   ['] move-SE-back  ,  
    ['] move-S-back   ,   ['] move-SW-back  ,    ['] move-W-back   ,   ['] move-NW-back  , 
    does> ( x y -- x' y' ) 
        get-heading cells + @ execute ;
create-back-motions   MOVE-BACK-BY-HEADING  ( x y -- x' y' )

\ === PIXEL LEVEL DISPLAY OF MOVEMENT
\ Paints after turtle has moved on when pen-down, erase if pen-down AND backspace, restore-pixel if pen-up
: cache-pixel   ( x y -- x y )        2dup lin peek-pixel    pixel pix!+ drop ;
: uncache-pixel ( -- pix1 pix2 )      pixel pix@+ drop ; 
: [paint-xy]    ( x y color -- x y )  pure-color >r  2dup lin BL  r>  [poke-pixel] ;  \ paint pixel with color
: restore-pixel ( x y -- x y )        2dup lin uncache-pixel [poke-pixel] ;           \ after turtle has moved on when pen-up

\ Every move or heading change makes turtle visible when cloaking is off; pixel is painted with color or white depending on fwd/bck
: move-fwd  ( x y -- x' y' )  is-penup? if restore-pixel else get-paint-color [paint-xy] then  move-fwd-by-heading  cloak-check ;
: move-back ( x y -- x' y' )  is-penup? if restore-pixel else white [paint-xy] then  move-back-by-heading cloak-check ;

\ === Displaying Turtle - show/hide use state transitions SET-HIDDEN and SET-VISIBLE
: change-heading ( 1/-1 -- )  get-heading + 8 mod set-heading cloak-check ;    \ nudges 45 degrees cw (positive) or ccw (negative)
: show-heading  ( x y h# -- )    \ draw symbol on screen where turtle is
    >r lin r> LOOKUP-HEADING-SYMBOL get-turtle-color get-paint-color fgbg [poke-pixel] ; 
: turtle-show       ( x y -- x y )  2dup get-heading show-heading set-visible ; 
: [turtle-hide]     ( x y -- x y )  restore-pixel ;   \ overwrite turtle at current location with cursor color
: turtle-hide       ( x y -- x y )  set-hidden  [turtle-hide] ;  \ hide and set state
: visibility-reset  ( x y -- x y )  is-visible? if turtle-show else turtle-hide then ;
: visibility-toggle ( x y -- x y )  is-visible? if turtle-hide else turtle-show then ; 

\ === Painting Text
: saveprompt       ( -- ) system-picture  screen-save ;
: unprompt         ( -- ) system-picture screen-restore ;
: display-text  ( x y asc -- x y )  \ text writes character-at-a-time in direction of turtle heading
    >r 2dup lin r>  get-turtle-color  get-paint-color  fgbg  [poke-pixel] move-fwd-by-heading cache-pixel ;
: textbox          ( c-addr nchar x y color1 color2 -- ) fgbg set-fgbg lin -rot write-string ;  \ red yellow is good
: promptbox        ( c-addr nchar -- ) 5 10 red lime textbox ;
: prompt-clear     ( -- c-addr nchar ) " Clear screen? (press y or ESC to clear, or any key to continue drawing.)" ;
: clear?           ( -- t/f )          prompt-clear saveprompt promptbox   'y' esc-key 2 key  check-keyset     unprompt ;
: prompt-quit      ( -- c-addr nchar ) " Quit? (press y or q to quit or any key to continue drawing)"  ;
: confirm-quit?    ( -- t/f )          prompt-quit  saveprompt promptbox     'y' 'q' 2 key  check-keyset     unprompt ;
: prompt-overwrite ( -- c-addr nchar ) " Shall I forget what you taught me before? (y/n)" ;
: overwrite?       ( -- t/f )          prompt-overwrite saveprompt promptbox     'y' 1 key check-keyset     unprompt ;
: prompt-full      ( -- c-addr nchar ) " Memory full! (500 instructions max). (Press any key...)" ;
: prompt-full2     ( -- c-addr nchar ) " ... try teaching the rest in the next skill.  (Press any key)" ;
: full! ( -- )  prompt-full   saveprompt promptbox key drop unprompt  
                prompt-full2  saveprompt promptbox key drop unprompt ;

\ === DIARY/RECORDING TEACHING/LEARNING:
\ Record strings
5       constant   num-recmsgs          \ number of different recording messages
12      constant   len-recmsg           \ msg length
11      constant   diary-numpos         \ char position in string for diarynum
create rec-state 1 bytes allot          \ state variables hold one of five states
   : rec-ready       0   rec-state c! ;
   : rec-recording   1   rec-state c! ;
   : rec-full        2   rec-state c! ;
   : rec-reminder    3   rec-state c! ;
   : rec-paused      4   rec-state c! ;
: get-rec-state ( -- r ) rec-state c@ ;
: create-recording-names   \ state of teaching/learning.  string dynamically updated (GET-REC-STR) shows which diary is active
    create ( -- ) here num-recmsgs len-recmsg * bytes allot
    " TeachSkill  "     attach+   " LEARNING    "     attach+
    " MemoryFull  "     attach+   " LEARNING    "     attach+
    " Paused      "     attach+
    drop
    does> ( idx base-adr -- c-addr nchar )      \ returns fixed width teach/learn state
        len-recmsg  rot get-fixedwidth-string ; 
create-recording-names  LOOKUP-RECMSG

\ ======================= KEY LOGGER / DIARY FUNCTION   For nested macros to work, a diary stack is needed
\ DIARIES
9 constant num-diaries                          \ number of distinct recordable macros
500 bytes constant diary-length                 \ max keypress codes in each macro  (frugality breeds resourcefulness, efficiency)
num-diaries diary-length *     constant diary-array-sz
create   diary    diary-array-sz bytes allot    \ allocate diary memory
diary diary-array-sz    0 fill                  \ zero out diary memory

\ DIARY MEMORY MANAGEMENT
\ OVERFLOW PROTECTION using DIARY-END array for each diary -- when end is reached FULL! message is shown to user and recording stops
create diary-end num-diaries cells allot        
diary-end num-diaries cells 0 fill                         \ zero out mem
: set-diary-end ( dx dptr -- ) swap cells diary-end + ! ;  \ store position 
: get-diary-end ( dx -- dptr ) cells diary-end + @ ;
: diary-full?  ( dx dptr -- t/f ) 2dup swap get-diary-end > ;

\ START POSITION -- start position for given diary
: get-diary-start ( dx -- dx dptr )  dup diary diary-length rot * + ;    

\ NEXT ENTRY LOCATIONS kept in DIARY-POS array for each diary -- next free place in each diary is also the stop point for playback
create diary-pos num-diaries cells allot  
diary-pos num-diaries cells 0 fill                          \ zero out mem
: set-diary-pos ( dx dptr -- )  swap cells diary-pos + ! ;  \ store position
: get-diary-pos ( dx -- dptr )  cells diary-pos + @ ;

\ ACTIVE INDEX -- holds which diary is active 1-9 (user numbers) 0-8 (system index)
create active-diary 1 bytes allot            
: set-diary-index ( dx -- ) active-diary c! ;
: get-diary-index ( -- dx ) active-diary c@ ;
: next-diary ( -- ) get-diary-index 1 + num-diaries mod  ( active-diary ) set-diary-index ;
\ conversions from user diary number (1-9) to coded diary index (0-8)
: diary-num2dx ( asc --  dx ) a2d 1 - ;       
: diary-dx2num ( dx  -- asc ) 1 + d2a ; 

\ over-write diary strings with dynamic diary number, displayed in menu
: add-diarynum  ( c-addr nchar -- c-addr nchar )    \ updates string with diary number
    2dup drop get-diary-index diary-dx2num swap diary-numpos + c! ;
: get-rec-str ( -- c-addr nchar )  get-rec-state  LOOKUP-RECMSG  add-diarynum  ;

\ INITIALIZING THE DIARIES AND ASSOCIATED STRUCTURES
: clear-diaries ( -- )
    0 get-diary-start num-diaries 0 do  ( dx dptr ) 
       2dup set-diary-pos                     \ set start
       diary-length + 2dup 1 - set-diary-end  \ set end
       ( dx dptr ) swap 1 + swap              \ next diary
    loop 2drop ;
: clear-diary ( dx -- ) get-diary-start 2dup set-diary-pos diary-length 0 fill drop ;

\ === VIEWING THE DIARY: we view diary in groups of octet instructions per row
8 constant octet
3 constant groups
octet groups * constant diary-row-size
: show-octet    ( dptr -- dptr' ) octet 0 do space c@+ emit loop ;
: show-row      ( dptr -- dptr' ) cr groups 0 do show-octet space space loop ; 
: cue-to-start  ( dx -- dptr )    diary diary-length rot * +  ;    \ cue to diary start position
: fetch-end     ( dx -- dptr )    cells diary-pos + @ ; 
: show-max-page ( nrows -- )   \ show screenful at a time
    0 do show-row i 1 + 20 mod 0 = if pause-key then loop ;  
: see-one-diary ( dx -- ) 
    dup cue-to-start swap fetch-end over - diary-row-size / 1 + show-max-page drop ;
: dump-diary    ( dx -- )  \ show entire diary, filled or not
    cue-to-start diary-length diary-row-size / show-max-page drop ;
: see-diaries ( -- )   \ pause after each diary view
    9 0 do cr ." Skill #" i dup diary-dx2num emit ." : " see-one-diary cr pause-esc if leave then loop cr ;
    
create diary-recording 1 cells allot   \ true if recording
: set-diary-recording   ( -- )     true  diary-recording c! ;
: clear-diary-recording ( -- )     false diary-recording c! ;
: is-recording?         ( -- t/f ) diary-recording c@ ;

\ === DIARY STACK - permits nested macros
create diary-stack 16 bytes allot               \ STACK - saves context for nested macros
create tods         1 cells allot               \ VARIABLE - holds top of diary stack
: set-tods ( dstack-ptr --  ) tods ! ; 
: get-tods ( -- dstack-ptr  ) tods @ ; 
: init-diary-stack diary-stack dup  16  $FF fill set-tods ;   \ set memory to $FF and set top of stack
init-diary-stack
: push-diary ( dx -- ) get-tods c!+ set-tods ;
: pop-diary  ( -- dx ) tods @ 1 - dup c@ swap set-tods ;

: dump-diary-stack ( -- ) diary-stack 16 0 do c@+ . loop drop ;
: see-diary-stack ( -- ) tods @ 1 + diary-stack do i c@ . loop ;
: test-diarystack ( 3 2 1 0 -- )   \ run it on command line
    dump-diary-stack see-diary-stack 
    3 2 1 0 push-diary push-diary push-diary push-diary 
    dump-diary see-diaries 
    pop-diary . pop-diary . pop-diary . ;

\ === MENU WRITING CODE ==
\ Draws static menu then overwrites it with dynamic content as it changes.
\ [SHOULD?] Do this to a virtual copy of the menu so that it persists even with screen changes (e.g. display help system, clear screen, etc.)
\ Text uses the default fgbg unless in a special mode, in which case this is provided instead.

\ DISPLAY COORDINATES  \ y-screen location of menus  (if change these, also change top-, bot-border-row)
0   constant menu0-row
22  constant menu1-row     
23  constant menu2-row     
24  constant menu3-row
\ x-screen coordinate of dynamic content in menu, indexed from 0
5   constant  h#   \ heading
10  constant  c#   \ color
4   constant  r#   \ recording status (teach/learn mode)
14  constant  t#   \ text mode
20  constant  d#   \ draw mode
58  constant  z#   \ cloaking mode
: menu0-text         " Turtle Logo Forth (v1.8) - [S/U]Save/Undo [Esc]ClearScr [#]Load [?]HELP [Q]UIT  " ;  \ top menu
: menu1-text         " Hdng:-- Mode:[Text ]Draw  [Arrows]MoveFwd,Turn L/R [Bkspc]Erase  [H]ome         " ;  \ bottom menus
: menu2-text         " [X/C]olor:------ [']Penup/dwn [;]Show/HideTurtle [Z]Cloak:Off [G]rid [W]ave     " ;
: menu3-text         " [T]:------------ [N]extSkill [1-9]Play-/[0]List-Skills                          " ;  \ [L]oop:--
\                    " ********************************************************************************"   \ template - max 80 char
\ menu text colors 
: menu-std-fgbg      ( -- ) white navy fgbg set-fgbg ;   \ white on navy
: menu-dynamic-fgbg  ( -- ) navy lime  fgbg set-fgbg ;   \ dynamic text
: menu-attn-fgbg     ( -- ) red lime   fgbg set-fgbg ;   \ attention mode
: menu-alert-fgbg    ( -- ) red yellow fgbg set-fgbg ;   \ alert mode
: menu-paint-fgbg   ( -- ) get-turtle-color get-paint-color fgbg set-fgbg ; \ turtle/text (fg) paint (bg) color
: create-rec-fgbg   \ lookup table
    create  \ fg  bg
        navy  c,  lime   c,    \ if rec-ready
        red   c,  yellow c,    \ if rec-recording
        navy  c,  yellow c,    \ if rec-full
        white c,  red    c,    \ if rec-reminder
        white c,  orange c,    \ if rec-paused
    does> ( -- fg bg )
       get-rec-state 2 * bytes + c@+ swap c@ fgbg set-fgbg ; 
create-rec-fgbg  menu-recstr-fgbg

\ static menu
: menu0 menu-std-fgbg  0 menu0-row lind   menu0-text  write-string ;
: menu1 menu-std-fgbg  0 menu1-row lind   menu1-text  write-string ;
: menu2 menu-std-fgbg  0 menu2-row lind   menu2-text  write-string ;
: menu3 menu-std-fgbg  0 menu3-row lind   menu3-text  write-string ;
\ over-write menu with dynamic text     \ Ex:  red yellow fgbg set-fgbg  " Hello!" 20 23 menu-overwrite ;
\ MENU1
: menu-show-heading  ( -- )  menu-dynamic-fgbg  h#  menu1-row lind  get-heading lookup-heading-string  write-string ;
: menu-show-mode            ( -- )  is-text-mode? 
    if    menu-attn-fgbg t# menu1-row lind " Text" write-string    menu-std-fgbg d# menu1-row lind  " Draw" write-string
    else  menu-attn-fgbg d# menu1-row lind " Draw" write-string    menu-std-fgbg t# menu1-row lind  " Text" write-string 
    then ;
\ MENU2
: menu-show-color    ( -- ) is-penup? 
       if menu-attn-fgbg else menu-paint-fgbg then  c#  menu2-row lind   get-pencolor-name  write-string ;
: menu-cloak-mode           ( -- )  is-cloaking-on? 
    if     menu-alert-fgbg  z# menu2-row lind  " On " write-string
    else   menu-attn-fgbg    z# menu2-row lind  " Off" write-string 
    then ;
\ MENU3
: menu-show-recording-state ( -- )  menu-recstr-fgbg  r#  menu3-row lind   get-rec-str  write-string ;
    
: paint-static-menus ( -- )   menu0 menu1 menu2 menu3   ;
: refresh-menus      ( -- )   \ over-write menus with dynamic portions
    menu-show-heading  menu-show-color  menu-show-recording-state  menu-show-mode  menu-cloak-mode ; 

: fresh-canvas       ( -- )   get-canvas-color   fill-screen   borders-draw   paint-static-menus  refresh-menus ;

\ === DIARY RECORDINGS
: show-recording ( -- )  set-diary-recording  rec-recording  menu-show-recording-state ;
\ ??? CODE-REVIEW - this changes recording state: flashes and SETS recording state.  Should flash and return to previous state.
: flash-recording ( -- )  rec-reminder  menu-show-recording-state 300 ms show-recording ;

: ok-to-write?     ( dx -- dx t/f ) \ checks if diary has content
    get-diary-start over get-diary-pos = if true else overwrite? then ;           
: start-recording  ( -- )  \ record macro from beginning if ok 
    get-diary-index  ok-to-write?  if clear-diary  show-recording else drop then  ;  
: resume-recording ( -- )  show-recording ;                                     \ record macro from previous position
: stop-recording   ( -- )  clear-diary-recording    rec-ready   menu-show-recording-state ;
: pause-recording  ( -- )  clear-diary-recording    rec-paused  menu-show-recording-state ; 

: warn-diary-full ( -- )  rec-full  menu-show-recording-state  full!  stop-recording ;
: diary-write-entry ( dx asc -- )  over get-diary-pos c!+ ( dx dptr' )  
    diary-full? if 2drop warn-diary-full else set-diary-pos then ;

\ ============ Control keys
\ Keyboard commands are associated with their ASCII codes.
\ Design is child friendly: upper & lower case letters are mapped to same command so program works even if caps lock gets pressed.
\    / is mapped to help (?) since ? requires use of SHIFT which is not easy for a child
\    (numbers are not affected by Caps Lock)
\ Note: added printable characters for all commands so can record in diary.  Arrows still work, but as synonymns.  Primary (recorded) commands are printable.
\ The primary key will get recorded into the diary, not all variations
'F'  constant  fwd-key      'f'  constant  fwd-key2      fwd-arrow  constant  fwd-key3       \ $C8=200d
'R'  constant  right-key    'r'  constant  right-key2    right-arrow  constant  right-key3    right-key  constant  cw-key \ 205d
'L'  constant  left-key     'l'  constant  left-key2     left-arrow   constant  left-key3     left-key   constant  ccw-key \ 203d
'B'  constant  erase-key    'b'  constant  erase-key2    backspace    constant    erase-key3
tick constant  penupdown-key
'#'  constant  hash-key
'c'  constant    color-fwd-key      'C'  constant    color-fwd-key2     \ change color
'x'  constant    color-back-key     'X'  constant    color-back-key2    \ change color
'n'  constant    diary-key          'N'  constant    diary-key2         \ RENAMED NEXT
't'  constant    record-key         'T'  constant    record-key2        \ RENAMED TEACH
'1'  constant    play-1-key    '2'  constant    play-2-key    '3'  constant    play-3-key    '4'  constant    play-4-key
'5'  constant    play-5-key    '6'  constant    play-6-key    '7'  constant    play-7-key    '8'  constant    play-8-key
'9'  constant    play-9-key
'0'  constant    show-diary-key
's'  constant    save-key          'S'  constant    save-key2
'u'  constant    undo-key          'U'  constant    undo-key2
';'  constant    show-hide-key      \ turtle hide/view 
'h'  constant    home-key          'H'  constant    home-key2
esc-key  constant    clear-screen-key
'?'  constant    help-key      '/'      constant    help-key2     \ Usability: in case a child does not know how to use shift key
'q'  constant    quit-key      'Q'      constant    quit-key2
'w'  constant    wave-key  'W'          constant    wave-key2  
'g'  constant    grid-key      'G'      constant    grid-key2
'['  constant    left-bracket
']'  constant    right-bracket
'z'  constant    cloak-key     'Z'      constant    cloak-key2

\ === KEY PROCESSING ENGINE ===
\ Turtle responds to user keypresses using a key mapping lookup table which associates every ascii key with the executable token (xt) 
\          giving its intended action (if it has one).
\ Mapping table is defined below (after all actions are defined) - search =keymap=
create   keymap   256 cells allot   \ creates storage for executable tokens (xt) to be logged at their ascii code offset location
: create-mapping  ( keymap <name> -- )   \ compiles keymap base address into <named> instruction.  
\                                          Generic design can be used for multi-language keymaps
    create ,                                 
    does> ( x y asc -- x y )                 \ key processing engine: key commands must consume ascii keycode and preserve x,y
        @ ( keymap)  over cells + @ ( x y asc xt ) execute ( x y )  refresh-menus  visibility-reset  ;
keymap create-mapping    LOOKUP-KEY-ACTION

\ === TURTLE PROCESSING COMMANDS (by LOOKUP-KEY-ACTION) ===
: is-current-diary? ( macro-asc -- macro-asc t/f ) dup diary-num2dx get-diary-index = ; 
: record-asc2diary ( asc -- asc ) get-diary-index over ( asc dx asc ) diary-write-entry ; 
: log-key-if-recording ( asc -- asc ) is-recording? if record-asc2diary then ;  \ required if command is to be logged

\ These commands are logged if recording is active ( normalized character key is logged, not necessarily the key that was pressed )
: penupdown ( x y asc -- x y ) log-key-if-recording  drop penup-toggle ;
: turtle-move-fwd  ( x y asc -- x' y' ) drop fwd-key   log-key-if-recording drop move-fwd  cache-pixel ;
: turtle-move-back ( x y asc -- x' y' ) drop erase-key log-key-if-recording drop move-back cache-pixel ;
: heading-change-left  ( asc -- )       drop left-key  log-key-if-recording drop -1 change-heading  ;  \ nudge 1 click ccw
: heading-change-right ( asc -- )       drop right-key log-key-if-recording drop  1 change-heading  ;  \ nudge 1 click cw
: color-change-fwd  ( x y asc -- x y )  \ nudge 1 color fwd.  \ changing color forces pendown mode  
    drop color-fwd-key  log-key-if-recording  drop 
    is-penup? if pen-down 1 change-color save-paint-color else 1 change-color then ;
: color-change-back ( x y asc -- x y ) 
    drop color-back-key  log-key-if-recording  drop 
    is-penup? if pen-down -1 change-color save-paint-color else -1 change-color then ;
: wave      ( x y asc -- x y )      \ show text on next row down
    log-key-if-recording drop 
    saveprompt
    2dup 1 + lin " Hello!" red yellow fgbg set-fgbg write-string 
    300 ms 
    unprompt ;     \ briefly flash hello

: print-location  ( x y -- x y )   2dup ." (x: " swap . ." y: " . ." )" ;    \ TO DO: Add this to status bar
: print-cartesian  ( x y -- x y )       \ calculate cartesian coordinates of turtle
    2dup ." (x: " swap . ." y: " user-rows 1 - swap - . ." )" ;

\ create and show a dynamic string
: grid-template  ( -- c-adr nchar )   " (x: --, y: --)" ;
: cartesian ( x y -- c-adr nchar )   \ create dynamic string 
    user-rows 1 - swap -  \ convert y to cartesian from matrix
    >r >r    \ save x y-cartesian
    grid-template swap dup 4 + r> ( get-x ) 2d2a rot c!+ c!+ 5 + r> ( get-y ) 2d2a rot c!+ c! swap ;  \ create string from template
: grid      ( x y asc -- x y )      \ briefly flash cartesian coordinates on next row down
    log-key-if-recording drop 
    saveprompt
    2dup 
    2dup 1 + lin -rot 
    cartesian red yellow fgbg set-fgbg write-string 
    500 ms 
    unprompt ;

\ These commands are SOMETIMES logged if recording is active.
: macro-playback ( x y asc -- x' y' )   \ show macro playing on screen
    diary-num2dx  dup set-diary-index  get-diary-start
    swap get-diary-pos swap
    2dup <= if 2drop  else  do i c@ LOOKUP-KEY-ACTION loop  then ;
: play-macro ( x y macro-asc -- x y )
    is-recording?  
if
    is-current-diary?    \ log key if not current diary
    \ nested macros allowed (mixture of primary and derived commands) but not recursion (calling oneself)
    if flash-recording  drop  \ recursion (calling onself) not allowed - warning and no playback 
    else 
       record-asc2diary                \ record macro-name not individual keypresses
       pause-recording                 \ playback individual keypresses but without recording       
       get-diary-index  push-diary     \ diary stack
       macro-playback                  \ want to see the macro playing on the screen as constructing the new macro
       pop-diary set-diary-index       \ store diary context from the stack
       resume-recording                \ allow recording to continue from same position
    then
else 
    macro-playback                  \ replay macro
then ;

\ REFERENCE POSITION
create xy-coord 2 bytes allot    \ storage for x y coord (used by undo)
: save-xy    ( x y -- x  y  )  2dup   swap xy-coord   c!+ c! ;
: restore-xy ( x y -- x' y' ) 2drop        xy-coord c@+ swap c@ ;    \ restore last saved x y

\ These commands are not logged (recording is flashed as reminder)
: save-point ( x y asc -- x y )                                   \ save screenshot hide turtle first to create a restore point
    drop is-recording? if flash-recording then save-xy 
        user-picture  screen-save  
        user-picture  fscr-save screen-store  ;
: restore-point ( x y asc -- x y )  \ restore to last saved image and replacing turtle coords
    drop is-recording? if flash-recording else user-picture screen-restore then restore-xy ; 
: undo  restore-point ;   \ synonymm
: load-file ( x y asc -- x y )  drop is-recording? if flash-recording then user-picture fscr-save screen-load restore-xy ;
: cloak        ( -- )  enable-cloaking   save-penup-state  pen-up   ;
: uncloak      ( -- )  disable-cloaking  restore-penup-state  penup-reset ; 
: cloak-toggle ( -- )  is-cloaking-on?  if  uncloak  turtle-show  else  cloak  turtle-hide then ;
: cloaking-onoff ( x y asc -- x y ) drop is-recording? if flash-recording else cloak-toggle then ;  \ toggle cloaking functionality
: turtle-showhide ( x y asc -- x y ) drop is-recording? if flash-recording else visibility-toggle then ;

\ These commands are not logged (no visible indication) and recording is stopped when they are pressed.
: [home] ( -- x y )  init-heading  init-turtle-position ;  \ first position
: turtle-home ( x y -- x y )  \ preserves turtle states, don't leave mark in last position & don't clobber home position
    is-penup? if restore-pixel else get-paint-color [paint-xy] then 2drop [home] cache-pixel ;
\ TODO: home to be recorded
: go-home ( x y asc -- x y ) drop is-recording? if stop-recording then turtle-home ;    \ reset turtle to home position
: recording-onoff ( x y asc -- x y )  drop is-recording? if stop-recording else start-recording then ;
: diary-next      ( x y asc -- x y )  drop is-recording? if stop-recording then next-diary ;

\ These commands are not allowed while recording
: is-quit? ( asc -- asc t/f ) 
        quit-key over <> over quit-key2 <> and  ( asc t/f )     \ if not q and not Q
        if false                                                \ then don't quit
        else is-recording? if flash-recording false             \ can't quit while recording
            else confirm-quit? then                             \ check user wants to quit - screen snapshot used by turtle-resume
        then ;
: clear-screen ( x y asc -- x y )  drop
        is-recording? if stop-recording then
        clear?   if  fresh-canvas  cache-pixel  then  ;

\ These commands are not logged but do not interrupt recording which resumes afterwards
: show-help   ( -- )             saveprompt  cr black-screen  welcome  pause-key  cr unprompt ; 
: help-screen ( x y asc -- x y ) drop show-help ;
: show-diary  ( x y asc -- x y ) drop saveprompt see-diaries unprompt ;

\  UNDECIDED logging behaviour
: text-mode-on  ( x y asc -- x y )  \ starting text mode brings pen down
    drop set-text-mode is-penup? if pen-down else no-action-needed then ;

\ === KEY PROCESSING ENGINE (part 2) ===
\ Actions (responses) have their executable tokens saved into KEYMAP at their ascii keycode offset.  
\     Non mapped keys are ignored by default
: add-key     ( xt keycode -- ) cells keymap + ! ;     \   stores response (xt) into keymap table at keycode offset location
: ignore-keypress  ( asc -- )   drop ;    
: default-ignore-allkeys  ( -- )    ['] ignore-keypress 255 0 do dup i add-key loop drop ;   \ keymap set with every key ignored 
: map-key-commands ( -- )   \ overlay specified keys with action/response ; rest will continue to ignore input by default
    default-ignore-allkeys       
        \ __xt__               __keycode__
  ['] penupdown                 penupdown-key      add-key
  ['] turtle-showhide           show-hide-key      add-key
  ['] text-mode-on              left-bracket       add-key    \ text mode turns off when detecting right-] or other exit keys
  ['] show-diary                show-diary-key     add-key
  ['] clear-screen              clear-screen-key   add-key
  ['] load-file                 hash-key           add-key
  ['] go-home               dup home-key           add-key       home-key2       add-key
  ['] diary-next            dup diary-key          add-key       diary-key2      add-key
  ['] recording-onoff       dup record-key         add-key       record-key2     add-key
  ['] save-point            dup save-key           add-key       save-key2       add-key
  ['] restore-point         dup undo-key           add-key       undo-key2       add-key
  ['] help-screen           dup help-key           add-key       help-key2       add-key
  ['] wave                  dup wave-key           add-key       wave-key2       add-key
  ['] grid                  dup grid-key           add-key       grid-key2       add-key
  ['] cloaking-onoff        dup cloak-key          add-key       cloak-key2      add-key
  ['] color-change-fwd      dup color-fwd-key      add-key       color-fwd-key2  add-key   
  ['] color-change-back     dup color-back-key     add-key       color-back-key2 add-key
  ['] turtle-move-fwd       dup fwd-key            add-key   dup fwd-key2        add-key       fwd-key3    add-key
  ['] turtle-move-back      dup erase-key          add-key   dup erase-key2      add-key       erase-key3  add-key
  ['] heading-change-left   dup left-key           add-key   dup left-key2       add-key       left-key3   add-key
  ['] heading-change-right  dup right-key          add-key   dup right-key2      add-key       right-key3  add-key
  ['] play-macro            dup play-1-key         add-key   dup play-2-key      add-key   dup play-3-key  add-key   
                            dup play-4-key         add-key   dup play-5-key      add-key   dup play-6-key  add-key   
                            dup play-7-key         add-key   dup play-8-key      add-key       play-9-key  add-key  ;

\ === MAIN PROGRAM ===
: turtle-init ( -- x y )   \ initialization code, starts with what was on the screen at quit time
    map-key-commands
    clear-diaries  0 set-diary-index
    disable-cloaking
    text-video-mode   \ standard 80x25 16-color text mode
    red set-paint-color save-paint-color
    silver set-border-color     white set-canvas-color   fresh-canvas
    stop-recording
    draw-mode  set-visible  pen-down
    [home]  turtle-home
    save-xy
    user-picture fscr-quit screen-load cache-pixel   \ load last screen
    paint-static-menus  refresh-menus  borders-draw  \ restores with latest menu system
    saveprompt 
    user-picture   screen-save
    turtle-show  ;
: stop-char?    ( asc -- asc t/f )  dup is_char? not over ']' = or ;   \ non-chars or ] signal stop
: process-text-loop ( x y -- x y ) 
       begin ( x y ) key stop-char? ( x y asc t/f )              \ fetch keys until stop-text is signalled 
        dup if over erase-key3 = 
             if  not   else   draw-mode   refresh-menus then     \ backspace stays in text mode, all other non-char keys signal exit
             >r LOOKUP-KEY-ACTION r>                             \ stop-chars are processed as turtle instructions
        else >r display-text r>
        then
       until ;
: process-key-loop ( x y -- x y )    \ fetch keys until quit, entering separate text processing loop if text mode is selected
    begin  key  ( asc )  is-quit?  if drop true  else  LOOKUP-KEY-ACTION
          is-text-mode? if process-text-loop then 
          false 
      then 
    until ;
: show-quit-message ( -- )  cr ." To restart, type TURTLE"  cr ." To return to Windows, type EXIT." cr ;
: turtle-quit    ( -- ) is-recording? if stop-recording  then  
    [turtle-hide]
    saveprompt  
    system-picture  fscr-quit screen-store  
    visibility-reset 
    show-quit-message 
    bye   \ comment out in dev mode.  Otherwise exit Forth to DOS.
    ;  
: turtle-run     ( x y -- x y )  process-key-loop        turtle-quit ;
: turtle-resume  ( x y -- x y )  unprompt  turtle-run  ; 
: turtle-go      ( -- x y )      show-help    turtle-init      turtle-run  ;    \ start application
\ === END PROGRAM ===
